@@ -27,6 +27,7 @@
 #include <fcntl.h>
 #include "getutent.h"
 
+#ifndef __FreeBSD__
 char *getdomainname(char *dname, size_t maxsize)
 {
     struct utsname st_uname;
@@ -55,6 +56,7 @@ char *getdomainname(char *dname, size_t maxsize)
 /*  printf ("domainname: %s\n", dname); */
     return dname;
 }
+#endif /* !__FreeBSD__ */
 
 static int utmp_fd = -1;
 
@@ -85,7 +87,7 @@ struct utmp *getutent()  /* returns next utmp file entry   */
     if (utmp_fd < 0)
 	setutent();
 
-    if ((readval = read(utmp_fd, &s_utmp, sizeof(s_utmp))) < sizeof(s_utmp))
+    if ((readval = read(utmp_fd, &s_utmp, sizeof(s_utmp))) < sizeof(s_utmp)) {
 	if (readval == 0)
 	    return NULL;
 	else if (readval < 0) {
@@ -95,8 +97,17 @@ struct utmp *getutent()  /* returns next utmp file entry   */
 	    fprintf(stderr, "Partial record in %s [%d bytes]\n", _PATH_UTMP, readval );
 	    exit(1);
 	}
+    }
 
     return &s_utmp;
 }
 #endif
  
+
+/*
+ * Local variables:
+ *  c-indent-level: 4
+ *  c-basic-offset: 4
+ *  tab-width: 8
+ * End:
+ */
