@@ -18,7 +18,11 @@
 #ifdef	BSD
 #include "getutent.h"
 #endif
+#ifndef __FreeBSD__
 #include <lastlog.h>
+#endif /* !__FreeBSD__ */
+#include <pwd.h>
+#include <sys/types.h>
 
 #include "privs.h"
 
@@ -26,14 +30,15 @@ int columns;
 int times_on;
 
 typedef struct {
-    char tty[20];
-    char locale[16];
+    char tty[UT_LINESIZE];
+    char locale[UT_HOSTSIZE];
     long ip_addr;
     time_t time;
     BOOL writable;
 } TTY_FROM;
 
-TTY_FROM tty_list[20];
+#define MAX_TTYS	64
+TTY_FROM tty_list[MAX_TTYS];
 static int uid, nuid, ngid; 
 
 /*
@@ -63,11 +68,11 @@ BOOL check_illegal(char *str, char *user)
     if ((lstat(str, (struct stat *) &statbuf)) < 0)
 	return(FALSE);
 
-    if (S_ISREG(statbuf.st_mode))
-	return(FALSE);
-    else
+    if (!S_ISREG(statbuf.st_mode)) {
 	syslog(LOG_NOTICE, "%s's file %s is NOT A REGULAR FILE.",
 		user, str);
+	return(FALSE);
+    }
 
     return(TRUE);
 }
@@ -93,7 +98,7 @@ void show_stuff(char *user, int t1, int t2)
 
     if (can_show) {
 	columns++;
-	sprintf(formatter, "%%s%%-%d.%ds", (36 - strlen(prog_config.p_strings[t2])),
+	snprintf(formatter, sizeof(formatter), "%%s%%-%d.%ds", (36 - strlen(prog_config.p_strings[t2])),
 		36);
 	printf(formatter, prog_config.p_strings[t2], user);
 	do_columns();
@@ -109,6 +114,9 @@ BOOL exist(char *filename)
 {
     FILE *f;
 
+    if (!filename)
+	return(FALSE);
+
     f = fopen(filename, "r");
     if (f) {
 	fclose(f);
@@ -123,9 +131,10 @@ BOOL exist(char *filename)
  * This simply shows specified data for each option that users would
  * normally expect.  (What else was I supposed to say?  :)
  */
-void show_pfile(int t1, int t2, int t3, char *dir, char *disp, BOOL send_ret)
+void show_pfile(uid_t uid, gid_t gid, int t1, int t2, int t3, char *dir, char *disp, BOOL send_ret)
 {
     BOOL can_show = FALSE;
+    FILE *file = NULL;
 
     if (local_finger) {
 	if (prog_config.local_config_bits2 & t1)
@@ -138,8 +147,8 @@ void show_pfile(int t1, int t2, int t3, char *dir, char *disp, BOOL send_ret)
     if (can_show) {
 	char d[128];
 
-	bzero(d, 128);
-	sprintf(d, "%s/%s", dir, disp);
+	memset(d, 0, 128);
+	snprintf(d, sizeof(d), "%s/%s", dir, disp);
 
 	if (!(prog_config.config_bits3 & SHOW_HEADERS_FE)) {
 	    printf("%s\n", prog_config.p_strings[t2]);
@@ -154,13 +163,8 @@ void show_pfile(int t1, int t2, int t3, char *dir, char *disp, BOOL send_ret)
 		fflush(stdout);
 	    }
 
-	    if (!check_illegal(d, "CFINGERD")) {
-		FILE *file;
-
-		file = fopen(d, "r");
-		display_file(file);
-		fclose(file);
-	    }
+	    if ((file = open_file(d)) != NULL)
+		display_file(uid, gid, file);
 	} else {
 	    if (!(prog_config.config_bits3 & SHOW_HEADERS_FE))
 		printf("%s\n", prog_config.p_strings[t3]);
@@ -179,9 +183,10 @@ void show_pfile(int t1, int t2, int t3, char *dir, char *disp, BOOL send_ret)
  * This simply shows specified data for each option that users would
  * normally expect.  (What else was I supposed to say?  :)
  */
-void show_pfile2(int t1, int t2, int t3, char *dir, char *disp, BOOL send_ret)
+void show_pfile2(uid_t uid, gid_t gid, int t1, int t2, int t3, char *dir, char *disp, BOOL send_ret)
 {
     BOOL can_show = FALSE;
+    FILE *file = NULL;
 
     if (local_finger) {
 	if (prog_config.local_config_bits3 & t1)
@@ -194,7 +199,7 @@ void show_pfile2(int t1, int t2, int t3, char *dir, char *disp, BOOL send_ret)
     if (can_show) {
 	char d[128];
 
-	sprintf(d, "%s/%s", dir, disp);
+	snprintf(d, sizeof(d), "%s/%s", dir, disp);
 
 	if (!(prog_config.config_bits3 & SHOW_HEADERS_FE)) {
 	    printf("%s\n", prog_config.p_strings[t2]);
@@ -209,13 +214,8 @@ void show_pfile2(int t1, int t2, int t3, char *dir, char *disp, BOOL send_ret)
 		fflush(stdout);
 	    }
 
-	    if (!check_illegal(d, "CFINGERD")) {
-		FILE *file;
-
-		file = fopen(d, "r");
-		display_file(file);
-		fclose(file);
-	    }
+	    if ((open_file (d)) != NULL)
+		display_file(uid, gid, file);
 	} else {
 	    if (!(prog_config.config_bits3 & SHOW_HEADERS_FE))
 		printf("%s\n", prog_config.p_strings[t3]);
@@ -259,11 +259,16 @@ void show_lasttime_on(char *user)
 		last_tty.ip_addr = 0;
 #else
 	    if (!strncmp(user, (char *) ut.ut_user, strlen(user))) {
-		strcpy(last_tty.tty, ut.ut_line);
+		strncpy(last_tty.tty, ut.ut_line, UT_LINESIZE - 1);
+		last_tty.tty[UT_LINESIZE - 1] = 0;
+#endif
+		strncpy(last_tty.locale, ut.ut_host, UT_HOSTSIZE - 1);
+		last_tty.locale[UT_HOSTSIZE - 1] = 0;
+#ifdef SUNOS
+		last_tty.ip_addr = 0;
+#else
+ 		last_tty.ip_addr = ut.ut_addr;
 #endif
-		ut.ut_host[16] = 0;
-		strcpy(last_tty.locale, ut.ut_host);
-		last_tty.ip_addr = ut.ut_addr;
 		last_tty.time = ut.ut_time;
 	    }
 	}
@@ -279,9 +284,10 @@ void show_lasttime_on(char *user)
 
 	if (read(fd, &lastlog, sizeof(lastlog)) == sizeof(lastlog)) {
 	    found = TRUE;
-	    strcpy(last_tty.tty, lastlog.ll_line);
-	    lastlog.ll_host[UT_HOSTSIZE - 1] = 0;
-	    strcpy(last_tty.locale, lastlog.ll_host);
+	    strncpy(last_tty.tty, lastlog.ll_line, UT_LINESIZE - 1);
+	    last_tty.tty[UT_LINESIZE - 1] = 0;
+	    strncpy(last_tty.locale, lastlog.ll_host, UT_HOSTSIZE - 1);
+	    last_tty.locale[UT_HOSTSIZE - 1] = 0;
 	    last_tty.time = lastlog.ll_time;
 	}
     }
@@ -298,7 +304,7 @@ void show_lasttime_on(char *user)
     }
 
     if (last_tty.time > 0L) {
-	strftime(formatted, 80, "%a %b %d %I:%M %p (%Z)",
+	strftime(formatted, 80, prog_config.ltime_format,
 	    localtime(&last_tty.time));
 	printf("Last seen %s", (char *) formatted);
     } else {
@@ -462,10 +468,10 @@ void show_times_on(void)
 	if (!(prog_config.config_bits2 & SHOW_STRICTFMT))
 	    SEND_RAW_RETURN;
 
-	for (i = 0; i < times_on; i++) {
+	for (i = 0; (i < times_on) && (i < MAX_TTYS) ; i++) {
 	    char formatted[80];
 
-	    strftime(formatted, 80, "%a %b %d %I:%M%p (%Z)",
+	    strftime(formatted, 80, prog_config.ltime_format,
 		    localtime(&tty_list[i].time));
 	    printf("On since %s on %s",
 		    formatted, (char *) tty_list[i].tty);
@@ -479,7 +485,7 @@ void show_times_on(void)
 			    (char *) tty_list[i].locale : "");
 
 	    if (!(tty_list[i].writable))
-		printf("[MSG-N]");
+		printf(" [MSG-N]");
 
 	    SEND_RAW_RETURN_NO_FLUSH;
 	}
@@ -512,33 +518,36 @@ void show_login_stats(char *user)
 		struct stat buf;
 		char devfile[80];
 
-#ifdef	BSD
-		ut->ut_line[UT_LINESIZE-1] = 0;
-#endif
+		if (times_on < MAX_TTYS) {
+		    strncpy(tty_list[times_on].tty, (char *) ut->ut_line, sizeof(tty_list[times_on].tty));
+		    strncpy(tty_list[times_on].tty, (char *) ut->ut_line, UT_LINESIZE - 1);
+		    tty_list[times_on].tty[UT_LINESIZE-1] = 0;
 
-		strcpy(tty_list[times_on].tty, (char *) ut->ut_line);
+		    sprintf(devfile, "/dev/%s", (char *) ut->ut_line);
 
-		sprintf(devfile, "/dev/%s", (char *) ut->ut_line);
+		    USER_PRIVS(nuid, ngid);
+		    stat(devfile, &buf);
+		    NOBODY_PRIVS;
 
-		USER_PRIVS(nuid, ngid);
-		stat(devfile, &buf);
-		NOBODY_PRIVS;
-
-		if ((buf.st_mode & S_IWGRP) && (buf.st_mode & S_IWOTH))
-		    tty_list[times_on].writable = TRUE;
-		else
-		    tty_list[times_on].writable = FALSE;
-
-#ifdef	BSD
-		ut->ut_host[UT_HOSTSIZE-1] = 0;
-		tty_list[times_on].ip_addr = 0;
+#ifdef	HAS_TTY_GROUP
+		    if (buf.st_mode & S_IWGRP)
 #else
-		ut->ut_host[16] = 0;
-		tty_list[times_on].ip_addr = ut->ut_addr;
+			if ((buf.st_mode & S_IWGRP) && (buf.st_mode & S_IWOTH))
+#endif
+			    tty_list[times_on].writable = TRUE;
+			else
+			    tty_list[times_on].writable = FALSE;
+
+#if	defined(BSD) || defined(SUNOS)
+		    tty_list[times_on].ip_addr = 0;
+#else
+		    tty_list[times_on].ip_addr = ut->ut_addr;
 #endif
 
-		strcpy(tty_list[times_on].locale, (char *) ut->ut_host);
-		tty_list[times_on].time = ut->ut_time;
+		    strncpy(tty_list[times_on].locale, (char *) ut->ut_host, UT_HOSTSIZE - 1);
+		    tty_list[times_on].locale[UT_HOSTSIZE - 1] = 0;
+		    tty_list[times_on].time = ut->ut_time;
+		}
 		times_on++;
 	    }
     }
@@ -566,7 +575,7 @@ void show_login_stats(char *user)
 void show_mail_stats(char *user)
 {
     struct stat buf;
-    char mailfile[80];
+    char mailfile[200];
     BOOL can_show = FALSE, can_show_date = FALSE;
 
     if (local_finger) {
@@ -588,8 +597,34 @@ void show_mail_stats(char *user)
     if (can_show) {
 	char formatted[80];
 	int fm;
+	char *cp, *xp, *yp;
+	struct passwd *pwent = NULL;
 
-	sprintf(mailfile, "%s/%s", MAIL_SPOOL_DIR, user);
+	/*
+	 *  Expand mailbox, $HOME and $USER are supported.
+	 */
+	memset (mailfile, 0, sizeof(mailfile));
+	for (cp=prog_config.mailbox_file,xp=mailfile;*cp&&strlen(mailfile)<sizeof(mailfile)-1;cp++) {
+	    if (*cp == '$') {
+		cp++;
+		if (*cp == '$') {
+		    *(xp++) = *cp;
+		} else if (!strncasecmp(cp,"USER", 4)) {
+		    for (yp=user;*yp&&strlen(mailfile)<sizeof(mailfile)-1;yp++)
+			*(xp++) = *yp;
+		    cp+=3;
+		} else if (!strncasecmp(cp,"HOME", 4)) {
+		    if (pwent == NULL) {
+			pwent = getpwnam (user);
+		    }
+		    if (pwent)
+			for (yp=pwent->pw_dir;*yp&&strlen(mailfile)<sizeof(mailfile)-1;yp++)
+			    *(xp++) = *yp;
+		    cp+=3;
+		}
+	    } else
+		*(xp++) = *cp;
+	}
 
 	USER_PRIVS(nuid, ngid);
 	fm = stat(mailfile, &buf);
@@ -597,10 +632,10 @@ void show_mail_stats(char *user)
 
 	if (fm != -1) {
 	    if (can_show_date)
-		strftime(formatted, 80, "%a %b %d %I:%M%p (%Z)",
+		strftime(formatted, 80, prog_config.ltime_format,
 		    localtime(&buf.st_atime));
 	    else
-		strftime(formatted, 80, "%I:%M %p (%Z)",
+		strftime(formatted, 80, prog_config.stime_format,
 		    localtime(&buf.st_atime));
 
 	    if ((int) buf.st_size == 0)
@@ -610,7 +645,7 @@ void show_mail_stats(char *user)
 		if (buf.st_mtime > buf.st_atime) {
 		    char formatted2[80];
 
-		    strftime(formatted2, 80, "%a %b %d %I:%M%p (%Z)",
+		    strftime(formatted2, 80, prog_config.ltime_format,
 			localtime(&buf.st_mtime));
 		    printf("This user has unread mail since %s\n",
 			(char *) formatted2);
@@ -619,8 +654,157 @@ void show_mail_stats(char *user)
 	} else
 	    printf("This user has no mail or mail spool.\n");
     }
+
 }
 
+  /*
+  * SHOW_QMAIL_STATS
+  *
+  * This shows whether the specified user has mail, and how long ago the
+  * user read mail.
+  */
+void show_qmail_stats(char *user, char *homedir, char *alias)
+    {
+	struct stat newbuf;
+	struct stat tmpbuf;
+	char mailfile[80];
+	BOOL can_show = FALSE, can_show_date = FALSE;
+	char *nomailmsg = "This user has no mail or mail spool.\n";
+ 
+	if (local_finger) {
+	    if (prog_config.local_config_bits1 & SHOW_LRMAIL)
+		can_show = TRUE;
+	} else {
+	    if (prog_config.config_bits1 & SHOW_LRMAIL)
+		can_show = TRUE;
+	}
+ 
+	if (local_finger) {
+	    if (prog_config.local_config_bits1 & SHOW_MRDATE)
+		can_show_date = TRUE;
+	} else {
+	    if (prog_config.config_bits1 & SHOW_MRDATE)
+		can_show_date = TRUE;
+	}
+ 
+	if (can_show) {
+	    char formatted[80];
+	    char line[80];
+	    int fm;
+	    FILE *qmf = NULL;
+ 
+	    /* open ~/.qmail-alias or ~/.qmail */
+	    if(alias != NULL)
+		snprintf(mailfile, sizeof(mailfile), "%s/.qmail-%s", homedir, alias);
+	    else
+		snprintf(mailfile, sizeof(mailfile), "%s/.qmail", homedir);
+
+	    USER_PRIVS(nuid, ngid);
+	    qmf = fopen(mailfile, "r");
+	    NOBODY_PRIVS;
+
+	    if(qmf == NULL) {
+		printf("No .qmail file.\n");
+		return;
+	    }
+	    line[0] = '\0';
+ 
+	    /* We get the first line starting with '.' or '/' */
+	    USER_PRIVS(nuid, ngid);
+	    while(qmf != NULL && fgets(line, sizeof(line), qmf)) {
+		if(line[0] != '.' && line[0] != '/')
+		    line[0] = '\0';
+		else
+		    break;
+	    }
+	    fclose(qmf);
+	    NOBODY_PRIVS;
+
+	    if(line[0] == '\0') {
+		printf("No local mail storage.\n");
+		return;
+	    }
+ 
+	    /* chdir() to home dir for paths relative to ~ */
+	    if(chdir(homedir) == -1)
+		return;
+ 
+	    if(line[strlen(line) - 1] == '\n')
+		line[strlen(line) - 1] = '\0';
+ 
+	    /* If the last char is not '/' then it's a sendmail format mail
+	       file and we use the old routines. */
+	    if(line[strlen(line) - 1] != '/') {
+		USER_PRIVS(nuid, ngid);
+		fm = stat(line, &newbuf);
+		NOBODY_PRIVS;
+
+		if (fm != -1) {
+		    if (can_show_date) {
+			strftime(formatted, 80, prog_config.ltime_format,
+				 localtime(&newbuf.st_atime));
+		    } else {
+			strftime(formatted, 80, prog_config.stime_format,
+				 localtime(&newbuf.st_atime));
+		    }
+ 
+		    if ((int) newbuf.st_size == 0)
+			printf("This user has no mail.\n");
+		    else {
+			printf("This user last read mail %s\n", (char *) formatted);
+			if (newbuf.st_mtime > newbuf.st_atime) {
+			    char formatted2[80];
+ 
+			    strftime(formatted2, 80, prog_config.ltime_format,
+				     localtime(&newbuf.st_mtime));
+			    printf("This user has unread mail since %s\n",
+				   (char *) formatted2);
+			}
+		    }
+		} else
+		    printf(nomailmsg);
+	    } else {
+		snprintf(mailfile, sizeof(mailfile), "%snew", line);
+		USER_PRIVS(nuid, ngid);
+		fm = stat(mailfile, &newbuf);
+		if(fm == -1) {
+		    printf(nomailmsg);
+		    return;
+		}
+		snprintf(mailfile, sizeof(mailfile), "%stmp", line);
+		fm = stat(mailfile, &tmpbuf);
+		if(fm == -1) {
+		    printf(nomailmsg);
+		    return;
+		}
+
+		NOBODY_PRIVS;
+ 
+		if (can_show_date) {
+		    strftime(formatted, 80, prog_config.ltime_format,
+			     localtime(&tmpbuf.st_atime));
+		} else {
+		    strftime(formatted, 80, prog_config.stime_format,
+			     localtime(&tmpbuf.st_atime));
+		}
+ 
+		if ((int) newbuf.st_size == 0)
+		    printf("This user has no mail.\n");
+		else {
+		    printf("This user last read mail %s\n", (char *) formatted);
+		    if (newbuf.st_mtime > tmpbuf.st_atime) {
+			char formatted2[80];
+ 
+			strftime(formatted2, 80, prog_config.ltime_format,
+				 localtime(&newbuf.st_mtime));
+			printf("This user has unread mail since %s\n",
+			       (char *) formatted2);
+		    }
+		}
+	    }
+	}
+    }
+
 /*
  * SHOW_FINGERINFO
  *
@@ -632,9 +816,28 @@ void show_fingerinfo(char *user)
     FILE *file;
     char username[80], room[80], work_phone[80], home_phone[80],
 	 other[80], fn[80];
+    BOOL qmail;
+    char acctname[100];
+    char *alias = NULL;
 
     columns = 0;
-    pwent = getpwnam(user);
+    qmail = ( strcasecmp(prog_config.mailbox_file, "QMAIL") == 0);
+
+    if(qmail) {
+        strncpy(acctname, user, sizeof(acctname) -1);
+        acctname[sizeof(acctname) -1] = '\0';
+        strtok(acctname, "-");
+	alias = strtok(NULL, "-");
+        pwent = getpwnam(acctname);
+    } else {
+        pwent = getpwnam(user);
+    }
+
+    /* Do a quick/dirty sanity check. */
+    if(pwent == NULL) {
+        show_bottom();
+        return;
+    }
 
     uid = pwent->pw_uid;
 
@@ -651,7 +854,14 @@ void show_fingerinfo(char *user)
     sscanf(pwent->pw_gecos, "%[^,],%[^,],%[^,],%[^,],%[^\r\n]\r\n",
 	username, room, work_phone, home_phone, other);
 
-    sprintf(fn, "%s/.nofinger", pwent->pw_dir);
+    if ((prog_config.no_finger_file != NULL) 
+	&& (strlen(pwent->pw_dir)+strlen(prog_config.no_finger_file)+1 < sizeof(fn)))
+        sprintf(fn, "%s/%s", pwent->pw_dir, prog_config.no_finger_file);
+    else
+      if (strlen(pwent->pw_dir)+10 < sizeof(fn))
+	sprintf(fn, "%s/.nofinger", pwent->pw_dir);
+      else
+	fn[0] = '\0';
 
     if (exist(fn)) {
 	if (!check_illegal((char *) fn, user)) {
@@ -663,8 +873,7 @@ void show_fingerinfo(char *user)
     USER_PRIVS(nuid, ngid);
     if ((file = fopen(fn, "r")) != NULL) {
 	NOBODY_PRIVS;
-	display_file(file);
-	fclose(file);
+	display_file(nuid, ngid, file);
     } else {
 	NOBODY_PRIVS;
 
@@ -688,29 +897,29 @@ void show_fingerinfo(char *user)
 	if (!(prog_config.config_bits2 & SHOW_STRICTFMT))
 	    SEND_RAW_RETURN;
 
-	show_mail_stats(user);
+        if(qmail)
+            show_qmail_stats(acctname, pwent->pw_dir, alias);
+        else
+	    show_mail_stats(user);
 	show_login_stats(user);
 
 	if (prog_config.config_bits2 & SHOW_STRICTFMT)
 	    SEND_RAW_RETURN;
 
-	/* I made a typo here ... thanks to stig@hackvan.com, it was
-	   corrected. */
-
-	show_pfile(SHOW_PROJECT, D_PROJECT, D_NO_PROJECT, pwent->pw_dir, 
+	show_pfile(nuid, ngid, SHOW_PROJECT, D_PROJECT, D_NO_PROJECT, pwent->pw_dir, 
 		prog_config.project_file, 
 	(prog_config.config_bits2 & SHOW_STRICTFMT) ? FALSE : TRUE);
 
-	show_pfile(SHOW_PLAN, D_PLAN, D_NO_PLAN, pwent->pw_dir,
+	show_pfile(nuid, ngid, SHOW_PLAN, D_PLAN, D_NO_PLAN, pwent->pw_dir,
 		prog_config.plan_file,
 	(prog_config.config_bits2 & SHOW_STRICTFMT) ? FALSE : TRUE);
 
 	/* Thanks to Andy Smith */
-	show_pfile(SHOW_PGPKEY, D_PGPKEY, D_NO_PGPKEY, pwent->pw_dir, 
+	show_pfile(nuid, ngid, SHOW_PGPKEY, D_PGPKEY, D_NO_PGPKEY, pwent->pw_dir, 
 		prog_config.pgpkey_file,
 	(prog_config.config_bits2 & SHOW_STRICTFMT) ? FALSE : TRUE);
 
-	show_pfile2(SHOW_XFACE, D_XFACE, D_NO_XFACE, pwent->pw_dir,
+	show_pfile2(nuid, ngid, SHOW_XFACE, D_XFACE, D_NO_XFACE, pwent->pw_dir,
 		prog_config.xface_file, FALSE);
     }
 }
@@ -725,6 +934,7 @@ void handle_standard(char *username)
 {
     char uname[600];
     int nums = 0;
+    char *buf;
 
     /*
     ** Support to change the "/W" or "-L" listings from Microsoft's
@@ -733,23 +943,33 @@ void handle_standard(char *username)
     **
     ** Ken Hollis 08/06/96
     */
-    if (username[0] == '/')
-	strcpy(username, username + 2);
-    else if (username[0] == '-')
-	strcpy(username, username + 2);
+    if (username[0] == '/' || username[0] == '-') {
+	nums=1;
+	if (username[2] == ' ')
+	    nums++;
+	strcpy(uname, username + nums);
+	strcpy(username, uname);
+    }
 
     if ((!(prog_config.config_bits2 & SHOW_FINGERFWD)) || 
 	check_exist(username))
 	show_top();
 
-    if (!check_exist(username)) {
+    if( !( strcmp(prog_config.mailbox_file, "QMAIL") == 0
+	&& check_exist_alias(username) )
+	&& !check_exist(username))
+    {
 	if (prog_config.config_bits2 & SHOW_FINGERFWD) {
 	    while(prog_config.forward[nums] != NULL) {
-		bzero(uname, 600);
+		memset(uname, 0, 600);
 		printf("\nChecking site %s for the specified user.\n", prog_config.forward[nums]);
 		fflush(stdout);
-		sprintf(uname, "finger %s\100%s | tail +2", username, prog_config.forward[nums]);
-		safe_exec(uname);
+		snprintf(uname, sizeof(uname), "/usr/bin/finger \"%s\100%s\" | /usr/bin/tail +2", username, prog_config.forward[nums]);
+		if ((buf = safe_exec(NOBODY_UID, NOBODY_GID, uname)) != NULL) {
+		    printf ("%s", buf);
+		    fflush(stdout);
+		    free (buf);
+		}
 		nums++;
 		log(LOG_USER, "Forwarded: ", username);
 	    }
@@ -766,3 +986,11 @@ void handle_standard(char *username)
 	show_bottom();
     }
 }
+
+/*
+ * Local variables:
+ *  c-indent-level: 4
+ *  c-basic-offset: 4
+ *  tab-width: 8
+ * End:
+ */
