@@ -39,6 +39,7 @@ void clear_cfingerd_vars(void)
     prog_config.pgpkey_file = NULL;
     prog_config.xface_file = NULL;
     prog_config.userlog_file = NULL;
+    prog_config.mailbox_file = NULL;
     prog_config.no_name_banner_file = NULL;
     prog_config.no_user_banner_file = NULL;
     prog_config.no_finger_file = NULL;
@@ -59,6 +60,9 @@ void clear_cfingerd_vars(void)
     for (i = 0; i < 24; i++)
 	prog_config.siglist[i] = NULL;
 
+    memset (prog_config.stime_format, 0, sizeof (prog_config.stime_format));
+    memset (prog_config.ltime_format, 0, sizeof (prog_config.ltime_format));
+
     for (i = 0; i < 40; i++) {
 	prog_config.fusers[i].user = NULL;
 	prog_config.fusers[i].script = NULL;
@@ -95,7 +99,7 @@ void start_cfingerd(void)
     int pid = getpid();
     char *proc;
 
-    proc = (char *) malloc(80);
+    proc = (char *) malloc(20);
 
     sprintf(proc, "cfingerd[%d]", pid);
     openlog(proc, LOG_NDELAY, LOG_DAEMON);
@@ -103,6 +107,23 @@ void start_cfingerd(void)
     clear_cfingerd_vars();
 }
 
+/*
+ * SET_TIME_FORMAT
+ *
+ * Sets two format strings for time formatting, according to clock24.
+ */
+void set_time_format()
+{
+    if (((local_finger) && (prog_config.local_config_bits3 & SHOW_CLOCK24)) ||
+	((!local_finger) && (prog_config.config_bits3 & SHOW_CLOCK24))) {
+	strncpy (prog_config.stime_format, "%H:%M (%Z)", sizeof (prog_config.stime_format));
+	strncpy (prog_config.ltime_format, "%a %b %d %H:%M (%Z)", sizeof (prog_config.ltime_format));
+    } else {
+	strncpy (prog_config.stime_format, "%I:%M %p (%Z)", sizeof (prog_config.stime_format));
+	strncpy (prog_config.ltime_format, "%a %b %d %I:%M %p (%Z)", sizeof (prog_config.ltime_format));
+    }
+}
+
 /*
  * CHECK_STATS
  *
@@ -125,13 +146,15 @@ void check_stats()
  * This routine simply parses a string given to the routine, and displays
  * information accordingly.
  */
-void parse_line(char *line)
+#define DISPLEN	160
+void parse_line(uid_t uid, gid_t gid, char *line)
 {
     int pos = 0, newpos = 0, done = FALSE, center_flag = FALSE, exec_line = FALSE;
     char command[80], *displine;
+    char *buf;
 
-    displine = (char *) malloc(160);
-    bzero(displine, 160);
+    displine = (char *) malloc(DISPLEN);
+    memset(displine, 0, DISPLEN);
 
     while(pos < strlen(line)) {
 	if (line[pos] != '$') {
@@ -139,9 +162,9 @@ void parse_line(char *line)
 		if (center_flag) {
 		    char allocation[2];
 
-		    bzero(allocation, 2);
-		    sprintf(allocation, "%c", line[pos]);
-		    displine = strcat(displine, (char *) allocation);
+		    memset(allocation, 0, 2);
+		    allocation[0] = line[pos];
+		    displine = strncat(displine, (char *) allocation, DISPLEN);
 		} else
 		    printf("%c", line[pos]);
 	    }
@@ -151,7 +174,7 @@ void parse_line(char *line)
 	    done = FALSE;
 	    newpos = 0;
 
-	    bzero(command, 80);
+	    memset(command, 0, 80);
 
 	    pos++;
 
@@ -169,7 +192,7 @@ void parse_line(char *line)
 
 		if (command[0] == '$') {
 		    if (center_flag)
-			displine = strcat(displine, "$");
+			displine = strncat(displine, "$", DISPLEN);
 		    else
 			printf("$");
 
@@ -182,11 +205,11 @@ void parse_line(char *line)
 			time_t tim = time(NULL);
 			char bleah[80];
 
-			strftime(bleah, 80, "%I:%M %p (%Z)",
+			strftime(bleah, 80, prog_config.stime_format,
 				localtime(&tim));
 
 			if (center_flag)
-			    displine = strcat(displine, (char *) bleah);
+			    displine = strncat(displine, (char *) bleah, DISPLEN);
 			else
 			    printf((char *) bleah);
 
@@ -202,7 +225,7 @@ void parse_line(char *line)
 				localtime(&tim));
 
 			if (center_flag)
-			    displine = strcat(displine, (char *) bleah);
+			    displine = strncat(displine, (char *) bleah, DISPLEN);
 			else
 			    printf((char *) bleah);
 
@@ -212,7 +235,7 @@ void parse_line(char *line)
 		    if ((!strncasecmp(command, "ident", 5)) &&
 			(strlen(command) == 5)) {
 			if (center_flag)
-			    displine = strcat(displine, ident_user);
+			    displine = strncat(displine, ident_user, DISPLEN);
 			else
 			    printf("%s", ident_user);
 
@@ -228,7 +251,7 @@ void parse_line(char *line)
 		    if ((!strncasecmp(command, "compile_datetime", 16)) &&
 			(strlen(command) == 16)) {
 			if (center_flag)
-			    displine = strcat(displine, (char *) COMPILE_DT);
+			    displine = strncat(displine, (char *) COMPILE_DT, DISPLEN);
 			else
 			    printf("%s", (char *) COMPILE_DT);
 
@@ -238,7 +261,7 @@ void parse_line(char *line)
 		    if ((!strncasecmp(command, "version", 7)) &&
 			(strlen(command) == 7)) {
 			if (center_flag)
-			    displine = strcat(displine, (char *) VERSION);
+			    displine = strncat(displine, (char *) VERSION, DISPLEN);
 			else
 			    printf("%s", (char *) VERSION);
 
@@ -248,29 +271,18 @@ void parse_line(char *line)
 		    if ((!strncasecmp(command, "exec", 4)) &&
 			(strlen(command) == 4) &&
 			(prog_config.config_bits3 & SHOW_EXEC)) {
-			FILE *pfile;
 			char crap[160];
 
-			bzero(crap, 160);
+			memset(crap, 0, 160);
 			sscanf(line, "$exec %[^\r\n]\r\n", crap);
 
 			exec_line = TRUE;
 
-			pfile = popen(crap, "r");
-			if (pfile) {
-			    char crap2[160];
-
-			    while(!feof(pfile)) {
-				bzero(crap2, 160);
-				fscanf(pfile, "%[^\r\n]\r\n", crap2);
-				printf("%s\n", crap2);
-				fflush(stdout);
-			    }
-			} else {
-			    printf("Exec of %s: %s\n", crap, strerror(errno));
+			if ((buf = safe_exec (uid, gid, crap)) != NULL) {
+			    printf ("%s", buf);
 			    fflush(stdout);
+			    free (buf);
 			}
-
 			done = TRUE;
 		    }
 		}
@@ -297,7 +309,7 @@ void parse_line(char *line)
  *
  * This displays the file specified.
  */
-void display_file(FILE *file_ent)
+void display_file(uid_t uid, gid_t gid, FILE *file_ent)
 {
     char line[128];
     int bsize = sizeof(line);
@@ -305,7 +317,7 @@ void display_file(FILE *file_ent)
     if (file_ent) {
 	while(fgets(line, bsize, file_ent)) {
 	    if (prog_config.config_bits3 & SHOW_PARSING)
-		parse_line(line);
+		parse_line(uid, gid, line);
 	    else
 		fputs(line, stdout);
 	}
@@ -333,7 +345,7 @@ void show_top(void)
     }
 
     if (can_show)
-	display_file(top_display);
+	display_file(NOBODY_UID, NOBODY_GID, top_display);
 }
 
 /*
@@ -349,6 +361,31 @@ BOOL check_exist(char *username)
     return((pwent != NULL) ? TRUE : FALSE);
 }
 
+/*
+ * CHECK_EXIST_ALIAS
+ *
+ * Check whether a Qmail user alias exists.
+ */
+BOOL check_exist_alias(char *username)
+{
+    struct passwd *pwent;
+    char acctname[100];
+
+    strncpy(acctname, username, 100);
+    acctname[sizeof(acctname) - 1] = '\0';
+    strtok(acctname, "-");
+
+    pwent = getpwnam(acctname);
+    if(pwent)
+    {
+        char qmfile[1024];
+        snprintf(qmfile, 1024, "%s/.qmail-%s", pwent->pw_dir
+                , &acctname[strlen(acctname) + 1]);
+        return exist(qmfile);
+    }
+    return FALSE;
+}
+
 /*
  * SHOW_NOTEXIST
  *
@@ -367,7 +404,7 @@ void show_notexist(void)
     }
 
     if (can_show)
-	display_file(nouser_display);
+	display_file(NOBODY_UID, NOBODY_GID, nouser_display);
 }
 
 /* 
@@ -388,7 +425,7 @@ void show_bottom(void)
     }
 
     if (can_show)
-	display_file(bottom_display);
+	display_file(NOBODY_UID, NOBODY_GID, bottom_display);
 }
 
 /*
@@ -428,7 +465,7 @@ char *get_localhost(void)
     ret = (char *) malloc(strlen((char *) hostname) +
 			  strlen((char *) domname) + 2);
 
-    sprintf(ret, "%s.%s", (char *) hostname, (char *) domname);
+    snprintf(ret, sizeof(ret), "%s.%s", (char *) hostname, (char *) domname);
     return(ret);
 }
 
@@ -441,12 +478,11 @@ char *get_localhost(void)
 void check_unknown(char *host)
 {
     if ((!(prog_config.config_bits2 & SHOW_NOBODY1413)) &&
-	(wildmat(ident_user, "unknown@*"))) {
+	!strncmp(ident_user,"unknown@", 8)) {
 	syslog(LOG_NOTICE, "Finger from %s %s",
 	    host, prog_config.p_strings[D_REJECT_HOST]);
-/*	if (prog_config.config_bits2 & SHOW_REJECTED)
-	    display_file(prog_config.rejected_file); */
-
+	if (prog_config.config_bits2 & SHOW_REJECTED)
+	    display_file(NOBODY_UID, NOBODY_GID, rejected_display);
 	log(LOG_IDENT, "Unknown not allowed from ", host);
 
 	exit(PROGRAM_OKAY);
@@ -498,8 +534,8 @@ void check_rejected(char *host)
 	    (wildmat(ip_address, prog_config.rejected[i]))) {
 	    syslog(LOG_NOTICE, "Finger from %s %s",
 		host, prog_config.p_strings[D_REJECT_HOST]);
-/*	    if (prog_config.config_bits2 & SHOW_REJECTED)
-		display_file(prog_config.rejected_file); */
+	    if (prog_config.config_bits2 & SHOW_REJECTED)
+		display_file(NOBODY_UID, NOBODY_GID, rejected_display);
 
 	    log(LOG_REJECTED, "Rejected host finger detected to ", host);
 
@@ -511,54 +547,95 @@ void check_rejected(char *host)
  * SAFE_EXEC
  *
  * This is a safe version of the "system", "exec" or "popen" command.  This
- * simply takes on the UID/GID of the NOBODY (or apparent NOBODY) UID/GID,
+ * simply takes on the UID/GID of the NOBODY (or apparent NOBODY UID/GID),
  * and runs the specified program as that user.  If it ever fails, it will
  * display in syslog the reason WHY it failed, along with a semi-detailed
  * message of the failure.
+ *
+ *
+ * After an exec() system call, the saved uid is lost. Therefore, a child
+ * program cannot use it to regain suid root permissions.
+ *
+ * W. Richard Stevens, Advanced Programming in the Unix Environment, page
+ * 213: "The saved set-user-ID is copied from the effective user ID by exec."
+ *
  */
-void safe_exec(char *cmd)
+char *safe_exec (uid_t uid, gid_t gid, char *cmd)
 {
     FILE *file;
     char line[600];
+    int fd[2];
+    int pid;
+    char *result = NULL;
+    unsigned long size = 0;
 
-#ifdef	NOBODY_UID
-#ifdef	NOBODY_GID
-    NOBODY_PRIVS;
-
-    file = popen(cmd, "r");
-
-    if (file) {
-	while(!feof(file)) {
-	    bzero(line, 600);
-	    fgets(line, 600, file);
-	    printf("%s", line);
-	    fflush(stdout);
-	}
-    } else {
-	syslog(LOG_ERR, "With the nobody UID/GID set,");
-	syslog(LOG_ERR, "safe_exec failed on %s: %s", cmd, strerror(errno));
+    if (pipe (fd) != 0) {
+	syslog(LOG_ERR, "pipe() failed, %s.", strerror(errno));
+	return NULL;
     }
 
-    RELINQUISH_PRIVS;
-#else
-    syslog(LOG_ERR, "No GID for Nobody defined.");
-#endif
-#else
-    syslog(LOG_ERR, "No UID for Nobody defined.");
-#endif
+    if ((pid = fork()) < 0) {
+	syslog(LOG_ERR, "fork() failed, %s.", strerror(errno));
+	return NULL;
+    }
+
+    if (pid > 0) {	/* father */
+	close (fd[1]);	/* for writing */
+
+	if ((result = (char *) malloc (1)) == NULL)
+	    return NULL;
+	*result = '\0';
+	memset (line, 0, 600);
+	while ((read (fd[0], line, sizeof (line)-1)) > 0) {
+	    if ((size += strlen (line)) < MAX_POPEN_BUF) {
+		syslog (LOG_INFO, "Reallocating %d bytes", size);
+		if ((result = (char *) realloc (result, size+1)) == NULL) {
+		    return NULL;
+		}
+		strcat (result, line);
+		memset (line, 0, 600);
+	    } else {	/* buffer limit reached */
+		close (fd[0]);
+		return result;
+	    }
+	}
+    } else {		/* child */
+	close (fd[0]);	/* for reading */
+	DROP_ALL_PRIVS(uid, gid);
+	file = popen(cmd, "r");
+
+	if (file) {
+	    while (!feof (file)) {
+		memset (line, 0, 600);
+		fgets (line, 600, file);
+		write (fd[1], line, strlen (line));
+	    }
+	} else {
+	    close (fd[0]);
+	    close (fd[1]);
+	    syslog(LOG_ERR, "safe_exec failed on %s: %s", cmd, strerror(errno));
+	}
+	close (fd[1]);
+	exit (0);
+    }
+
+    if (!strlen (result)) {
+	free (result);
+	result = NULL;
+    }
+    return result;
 }
 
 /*
- * MY_FOPEN
+ * OPEN_FILE
  *
  * Quick way to open files and check for errors at the same time.
  */
-FILE *my_fopen(char *filename)
+FILE *open_file(char *filename)
 {
-    FILE *myfile;
-
-    myfile = fopen(filename, "r");
-    return myfile;
+    if (check_illegal (filename, "CFINGERD"))
+	return (fopen (filename, "r"));
+    return NULL;
 }
 
 /*
@@ -569,8 +646,17 @@ FILE *my_fopen(char *filename)
  */
 void open_initial_files(void)
 {
-    top_display = my_fopen(prog_config.top_display_file);
-    bottom_display = my_fopen(prog_config.bottom_display_file);
-    noname_display = my_fopen(prog_config.no_name_banner_file);
-    nouser_display = my_fopen(prog_config.no_user_banner_file);
+    top_display = open_file (prog_config.top_display_file);
+    bottom_display = open_file (prog_config.bottom_display_file);
+    noname_display = open_file (prog_config.no_name_banner_file);
+    nouser_display = open_file (prog_config.no_user_banner_file);
+    rejected_display = open_file (prog_config.rejected_file);
 }
+
+/*
+ * Local variables:
+ *  c-indent-level: 4
+ *  c-basic-offset: 4
+ *  tab-width: 8
+ * End:
+ */
